--[[
TODO: [*]HIGH [+]Medium [x]Low [?]Idea
* Implement Vector<x,y> system. Sick of having 2 seperate functions for X and Y.

+ Collisions on ships when attacking planets. Right now they just sit there.
+ Fix rotations on ships to gradually turn to face destination
+ Basic AI
+ Fix Zoom On Cursor implementation
+ Get rid of Copy-Pasta entity system and write my own
+ Optimise Drawing calls. Ex. Don't highlight planets when the planet isn't visible
+ Teams
+ Make planets more pretty

X Figure out optimal quality for circles at any given scale factor

? Ships attack nearby ships
? Different ships with different stats
? Planets orbit stars and other objects ( May mess up gameplay for slow ships trying to follow orbit? )
--]]

local lgr = love.graphics

--Camera variables.
camX = 0
camY = 0

zoomScale = 0.3 -- Zoom level of the entire map.

--Holds the offsets to put the origin of the grid in the center of the screen
cX = lgr.getWidth( ) / 2
cY = lgr.getHeight() / 2

local angle = 0
game = {} -- Global Game table for different objects
game.planets = {}
game.planetTypes = {"gas", "metallic", "habitable", "water", "magma"} -- Not yet "used"

game.ships = {}
game.stars = {}   -- table which will hold our stars

game.selected = nil -- Eventually will be a table to hold objects selected. Right now its a single entity

function love.load()
	require("entities")
	require("resources")
	require("constants")

	-- Register our "Entities" in the game.
	ents.Register("planet")
	ents.Register("ship")
	
	lgr.setBackgroundColor( 0, 0, 0 )
	love.keyboard.setKeyRepeat(0.01, 0.2)-- Make it so the keys repeat when we hold them down.
	
	
	max_stars = 2000  -- how many stars we want
	local starfieldRadius = 9000 --math.max(lgr.getWidth(),lgr.getHeight())
	for i=1, max_stars do   -- generate the coords of our stars
		local x = math.random() * 2 * starfieldRadius - starfieldRadius
		local ylim = math.sqrt(starfieldRadius*starfieldRadius - x * x)
		local y = math.random() * 2 *ylim - ylim
		game.stars[i] = {x, y, (math.random() * 5), STAR_COLORS[math.random(1, #STAR_COLORS)]}   -- stick the values into the table
	end
   

	local planetRadi = 3000 -- Radius of where the planets can spawn
	
	--Make some planets
	for i = 0, 25 do
		local planetX = math.random() * 2 * planetRadi - planetRadi
		local ylim = math.sqrt(planetRadi*planetRadi - planetX * planetX)
		local planetY = math.random() * 2 *ylim - ylim
		local planet = ents.Create("planet",planetX, planetY)
		planet:setRadius( math.random(5,20) ) 
		planet:setPlanetType("habitable")
		planet:spawn()
		game.planets[i] = planet
	end
end

function drawStarfield(dt)
	--Draw the star field
	local stars = game.stars
	lgr.push() --Store origianl
	lgr.translate(cX-camX*zoomScale, cY+camY*zoomScale)
	lgr.setPointStyle( "rough" )	
    lgr.rotate(angle) -- Rotate the space to move stars
		
	for i=1, #stars do   -- loop through all of our stars
		local color = stars[i][4]
		lgr.setColor( color[1], color[2], color[3], 255 )
		lgr.setPointSize(stars[i][3]*zoomScale)
		lgr.point(stars[i][1] * zoomScale, stars[i][2] * zoomScale)   -- draw each point
	end
	lgr.pop() -- reload original
end

function drawPlanets(dt)
	lgr.push() -- Save Original
	lgr.translate(cX-camX*zoomScale, cY+camY*zoomScale)
	--Draw all of our planets
	for k, v in pairs(game.planets) do
		v:draw()
	end
	lgr.pop()
end

function drawDebug(dt)
	lgr.setColor(255,255,255,255)
	lgr.print( "FPS: " .. love.timer.getFPS( ), 16, 16, 0, 1, 1 )
	lgr.print( "Scale: " .. zoomScale, 16, 26, 0, 1, 1 )	
	lgr.print( "Num Ships: " .. #game.ships, 16, 36, 0, 1, 1 )	
	
	lgr.circle("line", cX,cY, 10, 360) -- Mark the center of the screen.
end

function drawShips(dt)
	lgr.push() -- Save Original
	lgr.translate(cX-camX*zoomScale, cY+camY*zoomScale)
	for k,v in pairs(game.ships) do
		v:draw(dt)
	end
	
	
	--Draw seletcted entities:
	if game.selected then
		local selected = game.selected
		if selected:isVisible() then
			local sx, sy = selected:getSize()
			
			lgr.setColor( 0, 255, 0, 255 )
			lgr.setLineWidth( 1 )
			lgr.setLineStyle("rough")
			lgr.rectangle( "line", (selected.x - sx - PLAN_PAD)*zoomScale, (-selected.y - sy - PLAN_PAD )*zoomScale, (2*PLAN_PAD + sx * 2)*zoomScale, (2*PLAN_PAD + sy * 2)*zoomScale )
		end
	end
	
	lgr.pop()
end

function drawUI(dt)
	
end

function love.draw(dt)

	
	lgr.push() --Store origianl
	-- All of my objects required me to do this in each function so why not save some power and do it first
	--lgr.translate(cX-camX*zoomScale, cY+camY*zoomScale)
	
	--Draw our game layout
	drawStarfield(dt)
	drawPlanets(dt)
	drawShips(dt)
	
	-- Revert to standard grid to draw UI
	lgr.pop()
	
	--Draw UI
	drawDebug(dt)

end

function love.update(dt)
	
	--Set angle of the star's rotations
    angle = (angle + dt * math.pi/300) % (2*math.pi)
	
	--Camera Movement
	local key = love.keyboard.isDown -- Apparently this saves time looking up the function in the table?
	if key("w") then
		camY = clamp(camY + 50/zoomScale*dt, -3000, 3000)
	elseif key("s") then
		camY = clamp(camY - 50/zoomScale*dt, -3000, 3000)
	end
	if key("a") then
		camX = clamp(camX - 50/zoomScale*dt, -3000, 3000)
	elseif key("d") then
		camX = clamp(camX + 50/zoomScale*dt, -3000, 3000)
	end
	
	for k,v in pairs(game.ships) do
		v:update(dt)
	end
	for k,v in pairs(game.planets) do
		v:update(dt)
	end
end

function love.focus(bool)
end

function love.keypressed( key, unicode )
	if key == "f5" then--restart
		love.filesystem.load("main.lua")()
		love.load()
	end
end

function love.keyreleased( key, unicode )
	
end

function love.mousepressed( x, y, button )

	-- Broken "Zoom In On Cursor" algorithm. Just zoom in on center for now.
	
	if button == "wd" then
		--camX = camX - (camX - (((love.mouse.getX()-cX) / zoomScale) + camX)) / (2)
		--camY = camY - (camY - (((cY - love.mouse.getY()) / zoomScale) + camY)) / (2)
		zoomScale = math.min(zoomScale + 0.1, 25)
	elseif button == "wu" then
		--camX = camX - (camX - (((love.mouse.getX()-cX) / zoomScale) + camX)) / (2)
		--camY = camY - (camY - (((cY - love.mouse.getY()) / zoomScale) + camY)) / (2)
		zoomScale = math.max(zoomScale - 0.1, 0.1)
	end
	
	-- Selection control Left-click to select, Right-click to attack.
	if button == "l" then
		game.selected = getObjectUnderCursor()
	elseif button == "r" then
		if game.selected then
			local target = getObjectUnderCursor()
			if target then
				print(target)
				game.selected:queueShips(target,20)
			end
		end
	end
end

function getObjectUnderCursor()
	for k, v in pairs(ents.objects) do
		if v.selectable and v:isVisible() and v:distance(getWorldX(), getWorldY()) <= math.min(v:getSize()) then
			return v
		end
	end
end

function love.mousereleased( x, y, button )
end

function love.quit()
end

function insideBox( px, py, x, y, wx, wy )
	if px > x and px < x + wx then
		if py > y and py < y + wy then
			return true
		end
	end
	return false
end

function clamp(value, min, max)
	if value > max then
		return max
	elseif value < min then
		return min
	else
		return value
	end
end

-- 2 mouse functions for getting where the mouse is in the world
function getWorldX()
	return ((love.mouse.getX()-cX) / zoomScale) + camX
end
function getWorldY()
	return ((cY - love.mouse.getY()) / zoomScale) + camY
end

function distance(x1,y1, x2,y2)
	return math.sqrt( (x2 - x1)^2 + (y2 - y1)^2 )
end